import yaml
import os


class ConfigBuilder:
    '''
    Generator for MVF configuration file.
    If a file already exists, user changes will not be overwritten.
    '''
    config = {
        'data': {
            'source': 'path_to_your_source_code',
            'lang': 'Python',
            'split': 'train_test',
            'test_size': 0.3,
            'input_features': [],
            'target_features': []
        },
        'models': [
            {
                'name': 'your_model_name',
                'lang': 'Python',
            },
        ],
    }

    def __init__(self, pth):
        '''
        Initialise the builder with template config.
        Updates with any existing config.

        Needs to be refactored. This method should not be creating
        a directory. This should be handled by mvf.cli.init
        '''
        # make project dir if not already exists
        if not os.path.exists(pth):
            os.makedirs(pth)
        # construct path to config file
        self.config_path = os.path.join(
            pth,
            'mvf_conf.yaml',
        )
        if os.path.isfile(self.config_path):
            with open(self.config_path, 'r') as f:
                existing_config = yaml.safe_load(f)
            self.config.update(existing_config)

    def write(self):
        '''
        Write config to file.
        '''
        with open(self.config_path, 'w') as f:
            yaml.safe_dump(self.config, f, default_flow_style=False)



from jinja2 import Environment, PackageLoader
import os
import jupytext


class FileGenerator:
    '''
    Generator for project files.

    Each method loads and renders a Jinja template and writes
    the template to a file.
    '''

    def __init__(self):
        '''
        Configure template file location for Jinja environment.
        '''
        self.env = Environment(
            loader=PackageLoader(
                'mvf',
                os.path.join('template', 'templates')
            )
        )

    def gen_preprocess(self, path, lang='Python'):
        '''
        Generate a data preprocessing template.
        '''
        template = self.env.get_template('preprocess.py')
        rendered = template.render(
            python=True if lang == 'Python' else False
        )
        # Convert to ipynb format
        nb = jupytext.reads(rendered, fmt="py:percent")
        out = jupytext.writes(nb, 'ipynb')
        with open(path, 'w') as f:
            f.write(out)

    def gen_models_py(self, models, path='models.py'):
        '''
        Generate a template Python models file.
        '''
        if models == []:
            pass
        else:
            template = self.env.get_template('models_py.py')
            rendered = template.render(
                models=models
            )
            with open(path, 'w') as f:
                f.write(rendered)

    def gen_models_r(self, models, path='models.R'):
        '''
        Generate a template R models file.
        '''
        if models == []:
            pass
        else:
            template = self.env.get_template('models_r.py')
            rendered = template.render(
                models=models
            )
            with open(path, 'w') as f:
                f.write(rendered)



# # Comparative Model Validation

# imports
from mvf.process import validation

# + tags=["parameters"]
upstream = None
product = None
models = []
split_type = ''
n_folds = 10
quantile_intervals = []
target_features = []
grouping_variable = None
temporal_variable = None
# -

# format variable
upstream = dict(upstream)

# ## Load predictions and validation data

# load validation data data from upstream
validation_data = validation.load_validation_data(
    upstream,
    split_type,
    n_folds
)

# load predictions from upstream
predictions = validation.load_predictions(
    upstream,
    models
)

# ## Error Metrics
# The error metrics reported are
# * MSE - Mean squared error
# * RMSE - Root mean squared error
# * MAE - Mean absolute error
# * MAPE - Mean absolute percentage error
# * R-squared - Coefficient of determination

validation.error_metrics(
    target_features,
    predictions,
    validation_data
)

# ## Quantile interval coverage
# For each of the intervals defined in `quantile_intervals`, what proportion of predicted values fall within the interval?

validation.qi_coverage(
    validation_data,
    target_features,
    predictions,
    quantile_intervals
)

# ## Model Sharpness
# How uncertain is the model? `sharpness` of a quantile interval is defined as the average width of the interval across all predictions.

validation.qi_sharpness(
    target_features, 
    predictions, 
    quantile_intervals
)

# ## Computational performance
# Record
# * the total size in bytes of the fit model(s).
# * the total time taken to fit the model(s).
# * the total time taken for the model(s) to predict.

validation.computational_performance(
    upstream,
    models,
    split_type,
    n_folds
)

# ## Plots

validation.Plot(
    models, 
    target_features, 
    validation_data, 
    upstream, 
    split_type
).plot(
    predictions,
    grouping_variable,
    quantile_intervals,
    temporal_variable
)


import os
from matplotlib import pyplot as plt
import feather
import pandas
from IPython.display import display, Markdown
import pickle
import sklearn.metrics


def load_validation_data(upstream, split_type, n_folds=10):
    '''
    Load validation data from upstream 'split_data' process.
    '''
    if split_type == 'train_test':
        validation_data = feather.read_dataframe(
            upstream['split_data']['test_y_data']
        ).reset_index(drop=True)

    elif split_type == 'k_fold':
        validation_data = []
        for i in range(1, n_folds+1):
            validation_data.append(
                feather.read_dataframe(
                    upstream['split_data'][f'fold_{i}_y_data'])
            )
        validation_data = pandas.concat(validation_data).reset_index(drop=True)
    return validation_data


def load_predictions(upstream, models):
    '''
    Load predictions from upstream 'predict_model' processes.
    '''
    predictions = {}
    for model_name in models:
        predictions[model_name] = feather.read_dataframe(
            upstream[f'{model_name}_predict']['predictions']
        ).reset_index(drop=True)
    return predictions


def error_metrics(target_features, predictions, validation_data):
    '''
    Calculate error metrics for each of the target features and for each model.
    '''
    # display a report for each target feature
    for target in target_features:
        # error
        error_df = pandas.DataFrame()
        for model, preds in predictions.items():
            # get predictions and ground truth
            mean_preds = preds[[target]].add_suffix('_pred', axis=1)
            ground_truth = validation_data[[
                target]].add_suffix('_truth', axis=1)
            # drop null validation values
            merged = pandas.concat([ground_truth, mean_preds], axis=1).dropna(
                subset=f'{target}_truth')
            # calculate metrics
            error_df.loc[model, 'MSE'] = sklearn.metrics.mean_squared_error(
                merged[f'{target}_truth'],
                merged[f'{target}_pred']
            )
            error_df.loc[model, 'RMSE'] = sklearn.metrics.mean_squared_error(
                merged[f'{target}_truth'],
                merged[f'{target}_pred'],
                squared=False
            )
            error_df.loc[model, 'MAE'] = sklearn.metrics.mean_absolute_error(
                merged[f'{target}_truth'],
                merged[f'{target}_pred']
            )
            error_df.loc[model, 'MAPE'] = sklearn.metrics.mean_absolute_percentage_error(
                merged[f'{target}_truth'],
                merged[f'{target}_pred']
            )
            error_df.loc[model, 'R-squared'] = sklearn.metrics.r2_score(
                merged[f'{target}_truth'],
                merged[f'{target}_pred']
            )
        print(f'Error metrics for `{target}` predictions:')
        display(error_df)


def qi_coverage(validation_data, target_features, predictions, quantile_intervals=[]):
    '''
    Calculate the coverage of for each quantile interval, for each target feature. 
    '''
    if quantile_intervals == []:
        print('No quantile intervals specified. Nothing to calculate.')
        return
    # display a report for each target feature
    for target in target_features:
        qi_coverage = pandas.DataFrame()
        # row index
        idx = 0
        for model, preds in predictions.items():
            for qi in quantile_intervals:
                lb = f'{target}_Q{min(qi) * 100:g}'
                ub = f'{target}_Q{max(qi) * 100:g}'
                if lb in preds.columns and ub in preds.columns:
                    qi_coverage.loc[idx, 'model'] = model
                    qi_coverage.loc[idx, 'interval'] = str(sorted(qi))
                    qi_coverage.loc[idx, 'level'] = max(qi) - min(qi)
                    # boolean mask for non-null validation values
                    m = validation_data[target].notna()
                    truth_above_lb = preds.loc[m,
                                               lb] <= validation_data.loc[m, target]
                    truth_below_ub = preds.loc[m,
                                               ub] >= validation_data.loc[m, target]
                    # boolean array for qi covering ground truth
                    in_interval = truth_above_lb & truth_below_ub
                    qi_coverage.loc[idx, 'coverage'] = in_interval.sum(
                    ) / in_interval.shape[0]
                    idx += 1
        print(f'Coverage metrics for `{target}` predictions:')
        display(qi_coverage.set_index(['model', 'interval']))


def qi_sharpness(target_features, predictions, quantile_intervals):
    '''
    Calculate the sharpness of each quantile interval for each target feature.
    '''
    if quantile_intervals == []:
        print('No quantile intervals specified. Nothing to calculate.')
        return
    # display a report for each target feature
    for target in target_features:
        qi_sharpness = pandas.DataFrame()
        for model, preds in predictions.items():
            for qi in quantile_intervals:
                lb = f'{target}_Q{min(qi) * 100:g}'
                ub = f'{target}_Q{max(qi) * 100:g}'
                if lb in preds.columns and ub in preds.columns:
                    qi_sharpness.loc[model, str(sorted(qi))] = (
                        preds[ub] - preds[lb]).mean()
        print(f'Sharpness metrics for `{target}` predictions:')
        display(qi_sharpness)


def computational_performance(upstream, models, split_type, n_folds=10):
    '''
    Calculates computational performance metrics for each model.
    '''
    # computational performance
    comp_perf = pandas.DataFrame()
    for model in models:
        # fit time
        with open(upstream[f'{model}_fit']['process_metadata'], 'rb') as f:
            fit_metadata = pickle.load(f)
        comp_perf.loc[model, 'fit time'] = fit_metadata[f'{model}_fit']
        # predict time
        with open(upstream[f'{model}_predict']['process_metadata'], 'rb') as f:
            predict_metadata = pickle.load(f)
        comp_perf.loc[model,
                      'predict time'] = predict_metadata[f'{model}_predict']
        # model size
        if split_type == 'train_test':
            # only one model
            comp_perf.loc[model, 'size'] = os.path.getsize(
                os.path.join(
                    'output',
                    'models',
                    f'{model}_fit'
                )
            )
        elif split_type == 'k_fold':
            # sum the model sizes
            total_size = 0
            for i in range(1, n_folds+1):
                total_size += os.path.getsize(
                    os.path.join(
                        'output',
                        'models',
                        f'{model}_fit_{i}'
                    )
                )
            comp_perf.loc[model, 'size'] = total_size
    display(comp_perf)


class Plot:
    '''
    Main plotting class. Holds matplotlib configs.
    '''
    colour_map = {
        0: 'tab:blue',
        1: 'tab:orange',
        2: 'tab:cyan',
        3: 'tab:purple',
        4: 'tab:grey'
    }
    line_map = {
        0: 'dotted',
        1: 'dashdot',
        2: 'dashed'
    }

    def __init__(self, models, target_features, validation_data, upstream, split_type):
        '''
        Sets basic class attributes.
        '''
        # set attributes
        self.models = models
        self.target_features = target_features
        self.validation_data = validation_data
        self.upstream = upstream
        self.split = split_type

    def plot(self, predictions, grouping_variable=None, quantile_intervals=None, temporal_variable=None):
        '''
        Main plotting method. Calls subsidiary plotting methods based on config.
        '''
        if grouping_variable and temporal_variable and self.split == 'train_test':
            self.group_var = grouping_variable
            self.temp_var = temporal_variable
            self.hierarchical_time_series_plot(predictions, quantile_intervals)
        else:
            print('Plotting for this project configuration has not been implemented yet.')

    def hierarchical_time_series_plot(self, predictions, quantile_intervals=None):
        '''
        Plots model predictions for hierarchical time series problems.
        '''
        if len(self.models) > 5:
            print(
                'Hierarchical time series plots are only supported for 5 models or fewer.')
            exit

        # extract upstream data
        X_train = feather.read_dataframe(
            self.upstream['split_data']['train_X_data']
        )
        X_test = feather.read_dataframe(
            self.upstream['split_data']['test_X_data']
        )
        y_train = feather.read_dataframe(
            self.upstream['split_data']['train_y_data']
        )
        y_test = feather.read_dataframe(
            self.upstream['split_data']['test_y_data']
        )
        train_data = pandas.concat([X_train, y_train], axis=1)
        test_data = pandas.concat([X_test, y_test], axis=1)

        if quantile_intervals and len(quantile_intervals) > 2:
            print('Only the first two quantile intervals will be included in the plots.')

        # create plots for each of the target features
        for target in self.target_features:
            display(Markdown(f'# Plots for {target} target feature'))

            # subset relevant ground truth data
            df = test_data[
                [self.group_var, self.temp_var, target]
            ].merge(
                train_data[
                    [self.group_var, self.temp_var, target]
                ],
                how='inner',
                on=[self.group_var, self.temp_var],
                suffixes=['_test', '_train']
            )
            # merge with predictions
            preds = [p.add_prefix(f'{m}_') for m, p in predictions.items()]
            df = pandas.concat([df] + preds, axis=1)

            # get min/max values to set common axis for all plots
            xmin = df[self.temp_var].min()
            xmax = df[self.temp_var].max()
            ymin = df[[f'{model}_{target}' for model in self.models]].min().min()
            ymax = df[[f'{model}_{target}' for model in self.models]].max().max()

            # create a plot for each instance of the grouping variable
            for var in df[self.group_var].unique():
                # subset data
                var_df = df.loc[df[self.group_var] == var]

                # add to figure for each model
                for c, model in enumerate(self.models):
                    # mean prediction line
                    plt.plot(
                        var_df[self.temp_var], var_df[f'{model}_{target}'], color=self.colour_map[c], label=f'{model}')
                    # quantiles
                    for l, qi in enumerate(quantile_intervals):
                        lb = min(qi)
                        ub = max(qi)
                        try:
                            plt.plot(
                                var_df[self.temp_var], var_df[f'{model}_{target}_Q{lb * 100:g}'], color=self.colour_map[c], linestyle=self.line_map[l])
                            plt.plot(
                                var_df[self.temp_var], var_df[f'{model}_{target}_Q{ub * 100:g}'], color=self.colour_map[c], linestyle=self.line_map[l])
                        except KeyError:
                            pass
                # ground truth points
                plt.scatter(var_df[self.temp_var],
                            var_df[f'{target}_test'], c='red')
                plt.scatter(var_df[self.temp_var],
                            var_df[f'{target}_train'], c='green')
                plt.xlim(xmin, xmax)
                plt.ylim(ymin, ymax)
                plt.xlabel(self.temp_var)
                plt.ylabel(target)
                plt.title(
                    f'{self.group_var}: {int(var)}, N train: {var_df[f"{target}_train"].notna().sum()}, N_test: {var_df[f"{target}_test"].notna().sum()}')
                plt.legend()
                plt.show()



from schema import Schema, Or, Optional, And


def check_config(config):
    '''
    Checks the given config against a schema.
    '''
    # schema for mvf_conf.yaml
    mvf_conf = Schema(
        {
            'data': {
                'source': And(
                    str,
                    lambda s: s.endswith('.ipynb'),
                    error='Invalid `source` file extension. Must be ipynb.'
                ),
                'lang': Or('Python', 'R'),
                'split': Or('train_test', 'k_fold'),
                'input_features': [str],
                'target_features': [str],
                Optional('grouping_variable'): str,
                Optional('test_size'): lambda x: 0 <= x <= 1,
                Optional('n_folds'): And(int, lambda x: x > 0),
                Optional('temporal_split'): {
                    'temporal_variable': str,
                    'test_size': Or(
                        lambda x: 0 <= x <= 1,
                        [lambda x: 0 <= x <= 1],
                    )
                }
            },
            Optional('output'): {
                Optional('quantile_intervals'): [
                    [lambda x: 0 <= x <= 1],
                ],
            },
            'models': [
                {
                    'name': str,
                    'lang': Or('Python', 'R'),
                    Optional('validation_step'): bool,
                    Optional('return_quantiles'): bool,
                }
            ],
        }
    )
    # validate given schema
    mvf_conf.validate(config)
    # additional validation
    if 'n_folds' in config['data'] and 'test_size' in config['data']:
        raise Exception(
            'Only one of `test_size` and `n_folds` may be specified.')
    if config['data']['split'] == 'train_test' and 'n_folds' in config['data']:
        raise Exception(
            '`n_folds` is not a valid parameter for a \'train_test\' split.')
    if config['data']['split'] == 'k_fold' and 'test_size' in config['data']:
        raise Exception(
            '`test_size` is not a valid parameter for a \'k_fold\' split.')
    if 'temporal_split' in config['data'] and 'grouping_variable' not in config['data']:
        raise Exception(
            'If the `temporal_split` parameters are provided, a grouping variable must also be provided.')
    if 'output' in config:
        if 'quantile_intervals' in config['output']:
            for pair in config['output']['quantile_intervals']:
                assert len(
                    pair) == 2, 'Two quantiles must be provided for each interval.'
            for model in config['models']:
                try:
                    assert isinstance(
                        model['return_quantiles'], bool), '`return_quantiles` must be Boolean type.'
                except KeyError:
                    raise Exception(
                        'If `quantile_intervals` are passed in `output`, the `return_quantiles` parameter must be passed for each model.')



import click
import os
import sys
from ploomber.exceptions import DAGBuildError
from mvf.cli import utils
from mvf.dag.builder import DagBuilder
from mvf.template.config import ConfigBuilder
from mvf.template.file_gen import FileGenerator
from rpy2.rinterface_lib.callbacks import logger as rpy2_logger
import logging
rpy2_logger.setLevel(logging.ERROR)


@click.group(
    help='''
    MVF

    See the documentation: 

        https://tomkimcta.gitlab.io/model-validation-framework
    '''
)
def mvf():
    '''
    CLI entry point.
    '''
    # Add working directory to PYTHONPATH to allow import of local modules
    sys.path.append(os.getcwd())


@click.command(
    help='''
        Initialise an MVF project. Default use will initialise the MVF
        project in the current working directory.
    '''
)
@click.option(
    '--directory',
    '-d',
    default='.',
    help='The directory in which to initialise the MVF project.'
)
def init(directory):
    '''
    Generates or updates an MVF configuration file.
    '''
    # check if default use
    if directory == '.':
        pth = os.getcwd()
        click.echo('Initialising MVF project in the current working directory')
    # path specified by user
    else:
        if os.path.isabs(directory):
            pth = directory
        # if relative path provided, construct absolute
        else:
            pth = os.path.join(
                os.getcwd(),
                directory
            )
        # check if the path provided by the user is already a file
        if os.path.isfile(pth):
            raise click.ClickException(
                'A file already exists at the location provided!')
        click.echo(f'Initialising MVF project in {pth}')
    # get confirmation from user
    click.confirm(f'Do you want to continue?', abort=True)
    cfg_builder = ConfigBuilder(pth)
    cfg_builder.write()
    # Print next steps
    click.echo('Next steps:\n')
    if directory != '.':
        click.secho(f'    cd {pth}', fg='green')
    click.echo('    Edit the mvf_conf.yaml file.\n')


@click.command(
    help='''
        Generate template files based on project configuration.
    '''
)
def scaffold():
    '''
    Generates template project files based on configuration.
    '''
    click.echo(f'This action will overwrite any existing source files.')
    click.confirm(f'Do you want to continue?', abort=True)
    # load project config
    config = utils.load_config(
        os.path.join(
            os.getcwd(),
            'mvf_conf.yaml'
        )
    )
    # parse specified models
    py_mods = [mod['name']
               for mod in config['models'] if mod['lang'] == 'Python']
    r_mods = [mod['name'] for mod in config['models'] if mod['lang'] == 'R']
    # generate files
    generator = FileGenerator()
    generator.gen_preprocess(config['data']['source'], config['data']['lang'])
    generator.gen_models_py(py_mods)
    generator.gen_models_r(r_mods)


@click.command(
    help='''
        Run an MVF project or process. Default use
        will trigger an incremental execution of the project.
    '''
)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help="Force execution of whole workflow."
)
@click.option(
    '--output',
    '-o',
    default='output',
    help='Specify the output directory.'
)
@click.option(
    '--process',
    '-p',
    default=None,
    help='Force execution of a specific process by name.'
)
def run(force, output, process):
    '''
    Runs an MVF project or process.
    '''
    # load project config
    config = utils.load_config(
        os.path.join(
            os.getcwd(),
            'mvf_conf.yaml'
        )
    )
    # build dag from config
    click.echo('Building MVF project workflow...')
    dag_builder = DagBuilder(config, output_dir=output)
    dag_builder.build()
    try:
        if process is None:
            # access the built dag and execute it
            click.echo('Running MVF project...')
            dag_builder.dag.build(force=force)
        else:
            # access a process and execute it
            dag_process = dag_builder.dag.get(process)
            dag_builder.dag.render(force=force)
            click.echo(f'Running {process} process...')
            dag_process.build(force=True)
    except DAGBuildError as e:
        click.echo(e)
    finally:
        # plot the dag
        click.echo('Plotting workflow...')
        dag_builder.dag.plot('project_workflow.html')


@click.command(
    help='''
        Plot an MVF workflow. If saving outputs to a
        custom directory, this directory should be passed as an
        option to get up-to-date process statuses.
    '''
)
@click.option(
    '--output',
    '-o',
    default='output',
    help='Specify the output directory.'
)
def plot(output):
    '''
    Plots an MVF workflow.
    '''
    # load project config
    config = utils.load_config(
        os.path.join(
            os.getcwd(),
            'mvf_conf.yaml'
        )
    )
    # build dag from config
    dag_builder = DagBuilder(config, output_dir=output)
    dag_builder.build()
    click.echo('Plotting workflow...')
    # access the built dag and plot it
    dag_builder.dag.plot('project_workflow.html')


@click.command(
    help='''
        Try out an example project.
    '''
)
@click.option(
    '--list',
    '-l',
    is_flag=True,
    default=False,
    help='List the available projects.'
)
@click.option(
    '--name',
    '-n',
    default=None,
    help='Name of project to download.'
)
@click.option(
    '--output',
    '-o',
    default=None,
    help='Target directory.'
)
def examples(list, name, output):
    '''
    Downloads an example project.
    '''
    # Use this list to manage which examples to expose
    examples_list = [
        'feature_scaling',
        'py_preprocess_train_test',
        'r_preprocess_kfold'
    ]
    if list:
        # List the examples in the console
        click.echo('The available examples are\n')
        for example_name in examples_list:
            click.echo(f'    {example_name}')
        click.echo('\nDownload them using `mvf examples -n name_of_project`')
    elif name is None:
        # Print the help text
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
    else:
        # Try download the example called name
        if name in examples_list:
            click.echo(f'Downloading {name} project...')
            # Download the example
            output_dir = output if output is not None else name
            utils.download_example(name, output_dir)
            # Print next steps
            click.echo('Next steps:\n')
            click.secho(f'    cd {output_dir}\n    mvf run\n', fg='green')
        else:
            click.echo(
                f'There is no project named {name}. Check the available projects with\n\n    mvf examples -l\n')


# add commands to group
mvf.add_command(init)
mvf.add_command(scaffold)
mvf.add_command(run)
mvf.add_command(plot)
mvf.add_command(examples)



# imports
import feather
import pandas
from mvf.process.utils import ProcessTimer, import_model_class, init_model, convert_to_r_data


def fit_model(product, upstream, lang, model_name, split_type, n_folds=10):
    '''
    Main method to fit model(s). Behaviour depends on project configuration.
    '''
    # start process timer
    timer = ProcessTimer(f'{model_name}_fit')

    # import model class
    model_module = import_model_class(lang)

    if split_type == 'train_test':
        # load training data
        X_train = feather.read_dataframe(upstream['split_data']['train_X_data'])
        y_train = feather.read_dataframe(upstream['split_data']['train_y_data'])

        # fit and save model
        fit(
            X_train,
            y_train,
            str(product['model']),
            model_module,
            lang,
            model_name
        )
    elif split_type == 'k_fold':
        # allocate memory
        X_data = []
        y_data = []
        for i in range(1, n_folds+1):
            # load data from upstream process
            X_data.append(
                feather.read_dataframe(
                    upstream['split_data'][f'fold_{i}_X_data']
                )
            )
            y_data.append(
                feather.read_dataframe(
                    upstream['split_data'][f'fold_{i}_y_data']
                )
            )
        
        # fit k models
        for i in range(1, n_folds+1):
            # get train set as all folds except i-th
            X_train = pandas.concat([x for j, x in enumerate(X_data) if j != i-1])
            y_train = pandas.concat([y for j, y in enumerate(y_data) if j != i-1])
            # fit and save model
            fit(
                X_train,
                y_train,
                str(product[f'model_{i}']),
                model_module,
                lang,
                model_name
            )
    # end process timer
    timer.end()
    # save process metadata
    timer.save(product['process_metadata'])


def fit(X_train, y_train, model_path, model_module, lang, model_name):
    '''
    Generic fit method.
    '''
    # if fitting an R model, must convert input data to R data.frame
    if lang == 'R':
        # convert pandas dataframe to R dataframe
        X_train = convert_to_r_data(X_train)
        y_train = convert_to_r_data(y_train)
    # initialise model
    model = init_model(
        model_module,
        lang,
        model_name
    )
    model.fit(X_train, y_train)
    model.save(model_path)


def fit_model_py(product: dict, params: dict) -> None:
    '''
    Run checks after fit process for Python models.
    '''
    import models
    # get model class
    model_class = getattr(models, params['model_name'])
    # different tests by split type
    if params['split_type'] == 'train_test':
        # load model
        model = model_class()
        model.load(product['model'])
        # check the model has a predict method
        check_loaded_model(
            model,
            product["model"]
        )
    elif params['split_type'] == 'k_fold':
        for i in range(1, params['n_folds'] + 1):
            # load model
            model = model_class()
            model.load(product[f'model_{i}'])
            # check the model has a predict method
            check_loaded_model(
                model,
                product[f'model_{i}']
            )
    else:
        raise NotImplementedError(
            f"The {params['split_type']} implementation is not tested. This is unacceptable.")


def fit_model_r(product: dict, params: dict) -> None:
    '''
    Run checks after fit process for R models.
    '''
    import rpy2.robjects as robjects
    import rpy2_r6.r6b as r6b
    # get model class
    r = robjects.r
    r.source('models.R')
    model_class = r6b.R6DynamicClassGenerator(r[params['model_name']])

    # different tests by split type
    if params['split_type'] == 'train_test':
        # load model
        model = model_class.new()
        model.load(str(product['model']))
        # check the model has a predict method
        check_loaded_model(
            model,
            product["model"]
        )
    elif params['split_type'] == 'k_fold':
        for i in range(1, params['n_folds'] + 1):
            # load model
            model = model_class.new()
            model.load(str(product[f'model_{i}']))
            # check the model has a predict method
            check_loaded_model(
                model,
                product[f'model_{i}']
            )
    else:
        raise NotImplementedError(
            f"The {params['split_type']} implementation is not tested. This is unacceptable.")


def check_loaded_model(model, path):
    '''
    Check a loaded model has a predict() method.
    '''
    assert hasattr(
        model, 'predict'), f'The model saved at {path} must have a predict() method.'
    assert callable(getattr(
        model, 'predict')), f'The model saved at {path} must have a predict() method. The class\' predict attribute is not currently callable.'



# Run on render of pipeline.
# Checks parameters of pipeline config are correct

from . import preprocess_data
from . import split_data
from . import fit_model
from . import validate_model
from . import predict_model
from . import validate


def validate_model_py(product, params):
    '''
    Run checks prior to validate process for Python model.
    '''
    import models
    # get model class
    model_class = getattr(models, params['model_name'])
    # init model
    model = model_class()
    # run common checks
    validate_model(product, params, model)


def validate_model_r(product, params):
    '''
    Run checks prior to validate process for R model.
    '''
    import rpy2.robjects as robjects
    import rpy2_r6.r6b as r6b
    # get model class
    r = robjects.r
    r.source('models.R')
    model_class = r6b.R6DynamicClassGenerator(r[params['model_name']])
    # init model
    model = model_class.new()
    # run common checks
    validate_model(product, params, model)

def validate_model(product, params, model):
    '''
    Run checks common to Python and R models.
    '''
    assert 'nb' in product, 'The \'nb\' product must be defined.'
    # check the model has a validate method
    assert hasattr(model, 'validate'), f'The model class defined for {params["model_name"]} must have a validate() method.'
    assert callable(getattr(model, 'validate')), f'The model saved at {params["model_name"]} must have a validate() method. The class\' load attribute is not currently callable.'



def preprocess_data(product: dict) -> None:
    '''
    Run checks prior to preprocess_data process.
    '''
    # check product params
    assert 'X_data' in product, 'The \'X_data\' product must be defined.'
    assert 'y_data' in product, 'The \'y_data\' product must be defined.'


import feather

def preprocess_data(product: dict, params):
    '''
    Run checks on dataset after preprocess_data process.
    '''
    # load datasets
    X = feather.read_dataframe(product['X_data'])
    y = feather.read_dataframe(product['y_data'])
    # check the expected features exist
    for input in params['input_features']:
        assert input in X.columns, f'{input} is specified as an input feature. A column named {input} must be in the \'X\' data frame.'
    for target in params['target_features']:
        assert target in y.columns, f'{target} is specified as an target feature. A column named {target} must be in the \'y\' data frame.'
    # check dimensionality of data
    assert len(X.shape) == 2, f'Expected the X data to have 2 dimensions, has {len(X.shape)}'
    assert len(y.shape) == 2, f'Expected the y data to have 2 dimensions, has {len(y.shape)}'  



# imports
import pandas
import feather
from sklearn.model_selection import KFold, GroupShuffleSplit, train_test_split
import numpy


def split_data(upstream, product, split_type, test_size=0.5, n_folds=10, grouping_variable=None, temporal_split=None):
    '''
    Main method to split dataset.
    '''
    # load data from upstream process
    X = feather.read_dataframe(upstream['preprocess_data']['X_data'])
    y = feather.read_dataframe(upstream['preprocess_data']['y_data'])

    if grouping_variable is None:
        groups = None
    else:
        groups = X[grouping_variable]

    # parse split_type parameter
    if split_type == 'train_test':
        if groups is not None:
            # split data
            gss = GroupShuffleSplit(
                n_splits=1,
                test_size=test_size,
            )
            for train_idx, test_idx in gss.split(X, y, groups):
                X_train = X.iloc[train_idx]
                y_train = y.iloc[train_idx]
                X_test = X.iloc[test_idx]
                y_test = y.iloc[test_idx]
                if temporal_split:
                    temp_var = temporal_split['temporal_variable']
                    # split target feature time series into train and test
                    y_test, y_train_sup = y_temporal_split(
                        X_test,
                        y_test,
                        temp_var,
                        grouping_variable,
                        temporal_split['test_size']
                    )
                    # add supplementary training data back into training set
                    X_train = pandas.concat([X_train, X_test])
                    y_train = pandas.concat([y_train, y_train_sup])
        else:
            X_train, X_test, y_train, y_test = train_test_split(
                X,
                y,
                test_size=test_size
            )
        # print metadata
        print(f'X_train shape: {X_train.shape}')
        print(f'X_test shape: {X_test.shape}')
        print(f'y_train shape: {y_train.shape}')
        print(f'y_test shape: {y_test.shape}')
        # save data for next process
        feather.write_dataframe(X_train, product['train_X_data'])
        feather.write_dataframe(X_test, product['test_X_data'])
        feather.write_dataframe(y_train, product['train_y_data'])
        feather.write_dataframe(y_test, product['test_y_data'])
    elif split_type == 'k_fold':
        kfolds = KFold(n_splits=n_folds)
        for i, (train_idx, test_idx) in enumerate(kfolds.split(X, groups=groups)):
            # split data
            fold_X_data = X.iloc[test_idx]
            fold_y_data = y.iloc[test_idx]
            # print metadata
            print(f'Fold {i+1} X shape: {fold_X_data.shape}')
            print(f'Fold {i+1} y shape: {fold_y_data.shape}')
            # save data for next process
            feather.write_dataframe(fold_X_data, product[f'fold_{i+1}_X_data'])
            feather.write_dataframe(fold_y_data, product[f'fold_{i+1}_y_data'])


def y_temporal_split(X_test, y_test, temp_var, grouping_variable, test_size):
    '''
    Temporal split of hierarchical time series.
    '''
    # allocate memory for supplementary training data
    y_train_sup = y_test.copy()

    target_not_null = y_test.notna().any(axis=1)
    df = X_test[target_not_null].groupby(
        grouping_variable
    )[temp_var].agg(
        ['min', 'max']
    )
    # determine the test size
    if isinstance(test_size, float):
        df['test_prop'] = numpy.full(
            df.shape[0], 
            test_size
        )
    else:
        df['test_prop'] = numpy.random.uniform(
            min(test_size),
            max(test_size),
            df.shape[0]
        )
    df['split_timestep'] = numpy.ceil(
        ((df['max'] - df['min']) * (1 - df.test_prop)) + df['min']
    )

    # mask to set test values to null
    m = X_test.groupby(
        grouping_variable
    ).apply(
        lambda g: g[temp_var] < df.loc[g[grouping_variable].unique()[
            0], 'split_timestep']
    ).reset_index(drop=True)
    # split time series
    y_test[m.values] = numpy.nan
    y_train_sup[~m.values] = numpy.nan

    return y_test, y_train_sup


# ---
# jupyter:
#   kernelspec:
{% if python -%}
#     display_name: Python 3
#     language: python
#     name: python3
{% else -%}
#     display_name: R
#     language: R
#     name: ir
{% endif %}# ---

# %%
# Imports
{% if python -%}
import pandas
import feather
{% else -%}
library(arrow)
{% endif %}

# %% tags=["parameters"]
# Do not edit this cell
{% if python -%}
upstream = None
product = None
input_features = []
target_features = []

print(f'Input features are: {input_features}')
print(f'Target features are: {target_features}')
{% else -%}
upstream = NULL
product = NULL
input_features = list()
target_features = list()

print('Input features:')
print(unlist(input_features))
print('Target features:')
print(unlist(target_features))
{% endif %}

# %% [markdown]
# Write code to fetch and preprocess data here.
#
# Ensure you split data into 2 {% if python -%}pandas.DataFrame{% else -%}R data.frame{% endif %} objects:
#
# * **X** - input features
# * **y** - target features

# %%
# Save data for next process
{% if python -%}
feather.write_dataframe(X, product['X_data'])
feather.write_dataframe(y, product['y_data'])
{% else -%}
write_feather(X, product[['X_data']])
write_feather(y, product[['y_data']])
{% endif %}